/*
  * Copyright 2010-2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License").
  * You may not use this file except in compliance with the License.
  * A copy of the License is located at
  *
  *  http://aws.amazon.com/apache2.0
  *
  * or in the "license" file accompanying this file. This file is distributed
  * on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
  * express or implied. See the License for the specific language governing
  * permissions and limitations under the License.
  */

#pragma once

#include <aws/core/Core_EXPORTS.h>

#include <aws/core/utils/memory/stl/AWSString.h>

namespace Aws
{
    namespace Http
    {
        /**
         * Enum representing URI scheme.
         */
        enum class Scheme
        {
            HTTP,
            HTTPS
        };

        namespace SchemeMapper
        {
            /**
             * Converts a Scheme instance to a String.
             */
            AWS_CORE_API const char* ToString(Scheme scheme);
            /**
            * Converts a string instance to a Scheme. Defaults to https.
            */
            AWS_CORE_API Scheme FromString(const char* name);
        } // namespace SchemeMapper
    } // namespace Http
} // namespace Aws

